@@ -1,4 +1,4 @@
-# Copyright 2002, 2003, 2004 Ben Escoto
+# Copyright 2002, 2003, 2004, 2005 Ben Escoto
 #
 # This file is part of rdiff-backup.
 #
@@ -24,7 +24,7 @@ import getopt, sys, re, os, cStringIO
 from log import Log, LoggerError, ErrorLog
 import Globals, Time, SetConnections, selection, robust, rpath, \
 	   manage, backup, connection, restore, FilenameMapping, \
-	   Security, Hardlink, regress, C, fs_abilities, statistics
+	   Security, Hardlink, regress, C, fs_abilities, statistics, compare
 
 
 action = None
@@ -58,10 +58,12 @@ def parse_cmdlineoptions(arglist):
 		return rpath.RPath(Globals.local_connection, path).normalize().path
 
 	try: optlist, args = getopt.getopt(arglist, "blr:sv:V",
-		 ["backup-mode", "calculate-average", "check-destination-dir",
-		  "compare", "compare-at-time=", "create-full-path",
-		  "current-time=", "exclude=", "exclude-device-files",
-		  "exclude-fifos", "exclude-filelist=",
+		 ["backup-mode", "calculate-average", "carbonfile",
+		  "check-destination-dir",
+		  "compare", "compare-at-time=", "compare-hash",
+		  "compare-hash-at-time=", "compare-full", "compare-full-at-time=",
+		  "create-full-path", "current-time=", "exclude=",
+		  "exclude-device-files", "exclude-fifos", "exclude-filelist=",
 		  "exclude-symbolic-links", "exclude-sockets",
 		  "exclude-filelist-stdin", "exclude-globbing-filelist=",
 		  "exclude-globbing-filelist-stdin", "exclude-mirror=",
@@ -82,7 +84,8 @@ def parse_cmdlineoptions(arglist):
 		  "remove-older-than=", "restore-as-of=", "restrict=",
 		  "restrict-read-only=", "restrict-update-only=", "server",
 		  "ssh-no-compression", "terminal-verbosity=", "test-server",
-		  "user-mapping-file=", "verbosity=", "version"])
+		  "user-mapping-file=", "verbosity=", "verify",
+		  "verify-at-time=", "version"])
 	except getopt.error, e:
 		commandline_error("Bad commandline options: %s" % str(e))
 
@@ -91,10 +94,12 @@ def parse_cmdlineoptions(arglist):
 		elif opt == "--calculate-average": action = "calculate-average"
 		elif opt == "--carbonfile": Globals.set("carbonfile_active", 1)
 		elif opt == "--check-destination-dir": action = "check-destination-dir"
-		elif opt == "--compare" or opt == "--compare-at-time":
-			action = "compare"
-			if opt == "--compare": restore_timestr = "now"
-			else: restore_timestr = arg
+		elif opt in ("--compare", "--compare-at-time",
+					 "--compare-hash", "--compare-hash-at-time",
+					 "--compare-full", "--compare-full-at-time"):
+			if opt[-8:] == "-at-time": restore_timestr, opt = arg, opt[:-8]
+			else: restore_timestr = "now"
+			action = opt[2:]
 		elif opt == "--create-full-path": create_full_path = 1
 		elif opt == "--current-time":
 			Globals.set_integer('current_time', arg)
@@ -187,10 +192,12 @@ def parse_cmdlineoptions(arglist):
 		elif opt == "--terminal-verbosity": Log.setterm_verbosity(arg)
 		elif opt == "--test-server": action = "test-server"
 		elif opt == "--user-mapping-file": user_mapping_filename = arg
+		elif opt == "-v" or opt == "--verbosity": Log.setverbosity(arg)
+		elif opt == "--verify": action, restore_timestr = "verify", "now"
+		elif opt == "--verify-at-time": action, restore_timestr = "verify", arg
 		elif opt == "-V" or opt == "--version":
 			print "rdiff-backup " + Globals.version
 			sys.exit(0)
-		elif opt == "-v" or opt == "--verbosity": Log.setverbosity(arg)
 		else: Log.FatalError("Unknown option %s" % opt)
 
 def check_action():
@@ -199,8 +206,10 @@ def check_action():
 	arg_action_dict = {0: ['server'],
 					   1: ['list-increments', 'list-increment-sizes',
 						   'remove-older-than', 'list-at-time',
-						   'list-changed-since', 'check-destination-dir'],
-					   2: ['backup', 'restore', 'restore-as-of', 'compare']}
+						   'list-changed-since', 'check-destination-dir',
+						   'verify'],
+					   2: ['backup', 'restore', 'restore-as-of',
+						   'compare', 'compare-hash', 'compare-full']}
 	l = len(args)
 	if l == 0 and action not in arg_action_dict[l]:
 		commandline_error("No arguments given")
@@ -263,7 +272,7 @@ def take_action(rps):
 	elif action == "backup": Backup(rps[0], rps[1])
 	elif action == "calculate-average": CalculateAverage(rps)
 	elif action == "check-destination-dir": CheckDest(rps[0])
-	elif action == "compare": Compare(*rps)
+	elif action.startswith("compare"): Compare(action, rps[0], rps[1])
 	elif action == "list-at-time": ListAtTime(rps[0])
 	elif action == "list-changed-since": ListChangedSince(rps[0])
 	elif action == "list-increments": ListIncrements(rps[0])
@@ -272,6 +281,7 @@ def take_action(rps):
 	elif action == "restore": Restore(*rps)
 	elif action == "restore-as-of": Restore(rps[0], rps[1], 1)
 	elif action == "test-server": SetConnections.TestConnections()
+	elif action == "verify": Verify(rps[0])
 	else: raise AssertionError("Unknown action " + action)
 
 def cleanup():
@@ -592,7 +602,7 @@ def restore_set_root(rpin):
 
 def ListIncrements(rp):
 	"""Print out a summary of the increments and their times"""
-	rp = require_root_set(rp)
+	rp = require_root_set(rp, 1)
 	restore_check_backup_dir(restore_root)
 	mirror_rp = restore_root.new_index(restore_index)
 	inc_rpath = Globals.rbdir.append_path('increments', restore_index)
@@ -602,24 +612,25 @@ def ListIncrements(rp):
 		print manage.describe_incs_parsable(incs, mirror_time, mirror_rp)
 	else: print manage.describe_incs_human(incs, mirror_time, mirror_rp)
 
-def require_root_set(rp):
+def require_root_set(rp, read_only):
 	"""Make sure rp is or is in a valid rdiff-backup dest directory.
 
-	Also initializes fs_abilities and quoting and return quoted rp if
-	necessary.
+	Also initializes fs_abilities (read or read/write) and quoting and
+	return quoted rp if necessary.
 
 	"""
 	if not restore_set_root(rp):
 		Log.FatalError(("Bad directory %s.\n" % (rp.path,)) +
 		  "It doesn't appear to be an rdiff-backup destination dir")
-	Globals.rbdir.conn.fs_abilities.single_set_globals(Globals.rbdir)
+	Globals.rbdir.conn.fs_abilities.single_set_globals(Globals.rbdir,
+													   read_only)
 	if Globals.chars_to_quote: return restore_init_quoting(rp)
 	else: return rp
 	
 
 def ListIncrementSizes(rp):
 	"""Print out a summary of the increments """
-	rp = require_root_set(rp)
+	rp = require_root_set(rp, 1)
 	print manage.ListIncrementSizes(restore_root, restore_index)
 
 
@@ -634,7 +645,7 @@ def CalculateAverage(rps):
 
 def RemoveOlderThan(rootrp):
 	"""Remove all increment files older than a certain time"""
-	rootrp = require_root_set(rootrp)
+	rootrp = require_root_set(rootrp, 0)
 	rot_require_rbdir_base(rootrp)
 	try: time = Time.genstrtotime(remove_older_than_string)
 	except Time.TimeException, exc: Log.FatalError(str(exc))
@@ -670,7 +681,7 @@ def rot_require_rbdir_base(rootrp):
 
 def ListChangedSince(rp):
 	"""List all the files under rp that have changed since restoretime"""
-	rp = require_root_set(rp)
+	rp = require_root_set(rp, 1)
 	try: rest_time = Time.genstrtotime(restore_timestr)
 	except Time.TimeException, exc: Log.FatalError(str(exc))
 	mirror_rp = restore_root.new_index(restore_index)
@@ -682,7 +693,7 @@ def ListChangedSince(rp):
 
 def ListAtTime(rp):
 	"""List files in archive under rp that are present at restoretime"""
-	rp = require_root_set(rp)
+	rp = require_root_set(rp, 1)
 	try: rest_time = Time.genstrtotime(restore_timestr)
 	except Time.TimeException, exc: Log.FatalError(str(exc))
 	mirror_rp = restore_root.new_index(restore_index)
@@ -691,7 +702,7 @@ def ListAtTime(rp):
 		print rorp.get_indexpath()
 	
 
-def Compare(src_rp, dest_rp, compare_time = None):
+def Compare(compare_type, src_rp, dest_rp, compare_time = None):
 	"""Compare metadata in src_rp with metadata of backup session
 
 	Prints to stdout whenever a file in the src_rp directory has
@@ -702,16 +713,32 @@ def Compare(src_rp, dest_rp, compare_time = None):
 
 	"""
 	global return_val
-	dest_rp = require_root_set(dest_rp)
+	dest_rp = require_root_set(dest_rp, 1)
 	if not compare_time:
 		try: compare_time = Time.genstrtotime(restore_timestr)
 		except Time.TimeException, exc: Log.FatalError(str(exc))
 
 	mirror_rp = restore_root.new_index(restore_index)
-	inc_rp = mirror_rp.append_path("increments", restore_index)
+	inc_rp = Globals.rbdir.append_path("increments", restore_index)
 	backup_set_select(src_rp) # Sets source rorp iterator
-	src_iter = src_rp.conn.backup.SourceStruct.get_source_select()
-	return_val = restore.Compare(src_iter, mirror_rp, inc_rp, compare_time)
+	if compare_type == "compare": compare_func = compare.Compare
+	elif compare_type == "compare-hash": compare_func = compare.Compare_hash
+	else:
+		assert compare_type == "compare-full", compare_type
+		compare_func = compare.Compare_full
+	return_val = compare_func(src_rp, mirror_rp, inc_rp, compare_time)
+
+def Verify(dest_rp, verify_time = None):
+	"""Check the hashes of the regular files against mirror_metadata"""
+	global return_val
+	dest_rp = require_root_set(dest_rp, 1)
+	if not verify_time:
+		try: verify_time = Time.genstrtotime(restore_timestr)
+		except Time.TimeException, exc: Log.FatalError(str(exc))
+
+	mirror_rp = restore_root.new_index(restore_index)
+	inc_rp = Globals.rbdir.append_path("increments", restore_index)
+	return_val = dest_rp.conn.compare.Verify(mirror_rp, inc_rp, verify_time)
 
 
 def CheckDest(dest_rp):
