@@ -103,6 +103,7 @@ class LowLevelPipeConnection(Connection):
 	b - string
 	q - quit signal
 	R - RPath
+	Q - QuotedRPath
 	r - RORPath only
 	c - PipeConnection object
 
@@ -127,6 +128,8 @@ class LowLevelPipeConnection(Connection):
 		log.Log.conn("sending", obj, req_num)
 		if type(obj) is types.StringType: self._putbuf(obj, req_num)
 		elif isinstance(obj, connection.Connection):self._putconn(obj, req_num)
+		elif isinstance(obj, FilenameMapping.QuotedRPath):
+			self._putqrpath(obj, req_num)
 		elif isinstance(obj, rpath.RPath): self._putrpath(obj, req_num)
 		elif isinstance(obj, rpath.RORPath): self._putrorpath(obj, req_num)
 		elif ((hasattr(obj, "read") or hasattr(obj, "write"))
@@ -150,8 +153,7 @@ class LowLevelPipeConnection(Connection):
 	def _putiter(self, iterator, req_num):
 		"""Put an iterator through the pipe"""
 		self._write("i",
-					str(VirtualFile.new(iterfile.RORPIterToFile(iterator))),
-					req_num)
+			str(VirtualFile.new(iterfile.MiscIterToFile(iterator))), req_num)
 
 	def _putrpath(self, rpath, req_num):
 		"""Put an rpath into the pipe
@@ -164,6 +166,12 @@ class LowLevelPipeConnection(Connection):
 					  rpath.index, rpath.data)
 		self._write("R", cPickle.dumps(rpath_repr, 1), req_num)
 
+	def _putqrpath(self, qrpath, req_num):
+		"""Put a quoted rpath into the pipe (similar to _putrpath above)"""
+		qrpath_repr = (qrpath.conn.conn_number, qrpath.base,
+					   qrpath.index, qrpath.data)
+		self._write("Q", cPickle.dumps(qrpath_repr, 1), req_num)
+
 	def _putrorpath(self, rorpath, req_num):
 		"""Put an rorpath into the pipe
 
@@ -232,9 +240,10 @@ class LowLevelPipeConnection(Connection):
 		elif format_string == "b": result = data
 		elif format_string == "f": result = VirtualFile(self, int(data))
 		elif format_string == "i":
-			result = iterfile.FileToRORPIter(VirtualFile(self, int(data)))
+			result = iterfile.FileToMiscIter(VirtualFile(self, int(data)))
 		elif format_string == "r": result = self._getrorpath(data)
 		elif format_string == "R": result = self._getrpath(data)
+		elif format_string == "Q": result = self._getqrpath(data)
 		else:
 			assert format_string == "c", header_string
 			result = Globals.connection_dict[int(data)]
@@ -252,6 +261,12 @@ class LowLevelPipeConnection(Connection):
 		return rpath.RPath(Globals.connection_dict[conn_number],
 						   base, index, data)
 
+	def _getqrpath(self, raw_qrpath_buf):
+		"""Return QuotedRPath object from raw buffer"""
+		conn_number, base, index, data = cPickle.loads(raw_qrpath_buf)
+		return FilenameMapping.QuotedRPath(
+			Globals.connection_dict[conn_number], base, index, data)
+
 	def _close(self):
 		"""Close the pipes associated with the connection"""
 		self.outpipe.close()
@@ -519,7 +534,7 @@ import Globals, Time, Rdiff, Hardlink, FilenameMapping, C, Security, \
 	   Main, rorpiter, selection, increment, statistics, manage, lazy, \
 	   iterfile, rpath, robust, restore, manage, backup, connection, \
 	   TempFile, SetConnections, librsync, log, regress, fs_abilities, \
-	   eas_acls, user_group
+	   eas_acls, user_group, compare
 
 Globals.local_connection = LocalConnection()
 Globals.connections.append(Globals.local_connection)
