@@ -22,13 +22,7 @@
 from __future__ import generators
 import tempfile, os, cStringIO
 import Globals, Time, Rdiff, Hardlink, rorpiter, selection, rpath, \
-	   log, static, robust, metadata, statistics, TempFile, eas_acls
-
-
-# This will be set to the time of the current mirror
-_mirror_time = None
-# This will be set to the exact time to restore to (not restore_to_time)
-_rest_time = None
+	   log, static, robust, metadata, statistics, TempFile, hash
 
 
 class RestoreError(Exception): pass
@@ -72,8 +66,8 @@ def ListChangedSince(mirror_rp, inc_rp, restore_to_time):
 	MirrorStruct.set_mirror_and_rest_times(restore_to_time)
 	MirrorStruct.initialize_rf_cache(mirror_rp, inc_rp)
 
-	old_iter = MirrorStruct.get_mirror_rorp_iter(_rest_time, 1)
-	cur_iter = MirrorStruct.get_mirror_rorp_iter(_mirror_time, 1)
+	old_iter = MirrorStruct.get_mirror_rorp_iter(MirrorStruct._rest_time, 1)
+	cur_iter = MirrorStruct.get_mirror_rorp_iter(MirrorStruct._mirror_time, 1)
 	collated = rorpiter.Collate2Iters(old_iter, cur_iter)
 	for old_rorp, cur_rorp in collated:
 		if not old_rorp: change = "new"
@@ -94,43 +88,23 @@ def ListAtTime(mirror_rp, inc_rp, time):
 	assert mirror_rp.conn is Globals.local_connection, "Run locally only"
 	MirrorStruct.set_mirror_and_rest_times(time)
 	MirrorStruct.initialize_rf_cache(mirror_rp, inc_rp)
-	old_iter = MirrorStruct.get_mirror_rorp_iter(_rest_time, 1)
+	old_iter = MirrorStruct.get_mirror_rorp_iter()
 	for rorp in old_iter: yield rorp
 	
-def Compare(src_iter, mirror_rp, inc_rp, compare_time):
-	"""Compares metadata in src_rp dir with metadata in mirror_rp at time"""
-	MirrorStruct.set_mirror_and_rest_times(compare_time)
-	MirrorStruct.initialize_rf_cache(mirror_rp, inc_rp)
-
-	mir_iter = MirrorStruct.get_mirror_rorp_iter(compare_time, 1)
-	collated = rorpiter.Collate2Iters(src_iter, mir_iter)
-	changed_files_found = 0
-	for src_rorp, mir_rorp in collated: 
-		if not mir_rorp: change = "new"
-		elif not src_rorp: change = "deleted"
-		elif src_rorp == mir_rorp: continue
-		else: change = "changed"
-		changed_files_found = 1
-		path_desc = (src_rorp and src_rorp.get_indexpath() or
-					 mir_rorp.get_indexpath())
-		log.Log("%-7s %s" % (change, path_desc), 2)
-		if change == "changed": # Log more description of difference
-			assert not src_rorp.equal_verbose_auto(mir_rorp, 3)
-
-	if not changed_files_found:
-		log.Log("No changes found.  Directory matches archive data.", 2)
-	MirrorStruct.close_rf_cache()
-	return changed_files_found
-
 
 class MirrorStruct:
 	"""Hold functions to be run on the mirror side"""
-	_select = None # If selection command line arguments given, use Select here
+	# If selection command line arguments given, use Select here
+	_select = None
+	# This will be set to the time of the current mirror
+	_mirror_time = None
+	# This will be set to the exact time to restore to (not restore_to_time)
+	_rest_time = None
+	
 	def set_mirror_and_rest_times(cls, restore_to_time):
-		"""Set global variabels _mirror_time and _rest_time on mirror conn"""
-		global _mirror_time, _rest_time
-		_mirror_time = cls.get_mirror_time()
-		_rest_time = cls.get_rest_time(restore_to_time)
+		"""Set class variabels _mirror_time and _rest_time on mirror conn"""
+		MirrorStruct._mirror_time = cls.get_mirror_time()
+		MirrorStruct._rest_time = cls.get_rest_time(restore_to_time)
 
 	def get_mirror_time(cls):
 		"""Return time (in seconds) of latest mirror"""
@@ -169,8 +143,8 @@ class MirrorStruct:
 
 		"""
 		# use dictionary to remove dups
-		if not _mirror_time: d = {cls.get_mirror_time(): None}
-		else: d = {_mirror_time: None}
+		if not cls._mirror_time: d = {cls.get_mirror_time(): None}
+		else: d = {cls._mirror_time: None}
 		if not rp or not rp.index: rp = Globals.rbdir.append("increments")
 		for inc in get_inclist(rp): d[inc.getinctime()] = None
 		for inc in get_inclist(Globals.rbdir.append("mirror_metadata")):
@@ -201,11 +175,11 @@ class MirrorStruct:
 		unwanted files from the metadata_iter.
 
 		"""
-		if rest_time is None: rest_time = _rest_time
+		if rest_time is None: rest_time = cls._rest_time
 
-		rorp_iter = eas_acls.GetCombinedMetadataIter(
-			Globals.rbdir, rest_time, restrict_index = cls.mirror_base.index,
-			acls = Globals.acls_active, eas = Globals.eas_active)
+		metadata.SetManager()
+		rorp_iter = metadata.ManagerObj.GetAtTime(rest_time,
+												  cls.mirror_base.index)
 		if not rorp_iter:
 			if require_metadata:
 				log.Log.FatalError("Mirror metadata not found")
@@ -282,7 +256,8 @@ class MirrorStruct:
 			mir_rorp.flaglinked(Hardlink.get_link_index(mir_rorp))
 		elif mir_rorp.isreg():
 			expanded_index = cls.mirror_base.index + mir_rorp.index
-			mir_rorp.setfile(cls.rf_cache.get_fp(expanded_index))
+			file_fp = cls.rf_cache.get_fp(expanded_index)
+			mir_rorp.setfile(hash.FileWrapper(file_fp))
 		mir_rorp.set_attached_filetype('snapshot')
 		return mir_rorp
 
@@ -371,7 +346,7 @@ class CachedRF:
 		rf = self.get_rf(index)
 		if not rf:
 			log.Log("""Error: Unable to retrieve data for file %s!
-The cause is probably data loss from the destination directory.""" %
+The cause is probably data loss from the backup repository.""" %
 					(index and "/".join(index) or '.',), 2)
 			return cStringIO.StringIO('')
 		return self.get_rf(index).get_restore_fp()
@@ -434,7 +409,8 @@ class RestoreFile:
 		"""
 		self.mirror_rp.inc_type = 'snapshot'
 		self.mirror_rp.inc_compressed = 0
-		if not self.inc_list or _rest_time >= _mirror_time:
+		if (not self.inc_list or
+			MirrorStruct._rest_time >= MirrorStruct._mirror_time):
 			self.relevant_incs = [self.mirror_rp]
 			return
 
@@ -461,7 +437,7 @@ class RestoreFile:
 		incpairs = []
 		for inc in self.inc_list:
 			time = inc.getinctime()
-			if time >= _rest_time: incpairs.append((time, inc))
+			if time >= MirrorStruct._rest_time: incpairs.append((time, inc))
 		incpairs.sort()
 		return [pair[1] for pair in incpairs]
 
@@ -484,6 +460,23 @@ class RestoreFile:
 
 	def get_restore_fp(self):
 		"""Return file object of restored data"""
+		def get_fp():
+			current_fp = self.get_first_fp()
+			for inc_diff in self.relevant_incs[1:]:
+				log.Log("Applying patch %s" % (inc_diff.get_indexpath(),), 7)
+				assert inc_diff.getinctype() == 'diff'
+				delta_fp = inc_diff.open("rb", inc_diff.isinccompressed())
+				new_fp = tempfile.TemporaryFile()
+				Rdiff.write_patched_fp(current_fp, delta_fp, new_fp)
+				new_fp.seek(0)
+				current_fp = new_fp
+			return current_fp
+
+		def error_handler(exc):
+			log.Log("Error reading %s, substituting empty file." %
+					(self.mirror_rp.path,), 2)
+			return cStringIO.StringIO('')
+
 		if not self.relevant_incs[-1].isreg():
 			log.Log("""Warning: Could not restore file %s!
 
@@ -494,16 +487,7 @@ created.  This error is probably caused by data loss in the
 rdiff-backup destination directory, or a bug in rdiff-backup""" %
 					(self.mirror_rp.path, self.relevant_incs[-1].lstat()), 2)
 			return cStringIO.StringIO('')
-		current_fp = self.get_first_fp()
-		for inc_diff in self.relevant_incs[1:]:
-			log.Log("Applying patch %s" % (inc_diff.get_indexpath(),), 7)
-			assert inc_diff.getinctype() == 'diff'
-			delta_fp = inc_diff.open("rb", inc_diff.isinccompressed())
-			new_fp = tempfile.TemporaryFile()
-			Rdiff.write_patched_fp(current_fp, delta_fp, new_fp)
-			new_fp.seek(0)
-			current_fp = new_fp
-		return current_fp
+		return robust.check_common_error(error_handler, get_fp)
 
 	def get_first_fp(self):
 		"""Return first file object from relevant inc list"""
@@ -633,15 +617,31 @@ class PatchITRB(rorpiter.ITRBranch):
 		self.patch_to_temp(rp, diff_rorp, tf)
 		rpath.rename(tf, rp)
 
+	def check_hash(self, copy_report, diff_rorp):
+		"""Check the hash in the copy_report with hash in diff_rorp"""
+		if not diff_rorp.isreg(): return
+		if not diff_rorp.has_sha1():
+			log.Log("Hash for %s missing, cannot check" %
+					(diff_rorp.get_indexpath()), 2)
+		elif copy_report.sha1_digest == diff_rorp.get_sha1():
+			log.Log("Hash %s of %s verified" %
+					(diff_rorp.get_sha1(), diff_rorp.get_indexpath()), 6)
+		else:
+			log.Log("Warning: Hash %s of %s\ndoesn't match recorded hash %s!"
+					% (copy_report.sha1_digest, diff_rorp.get_indexpath(),
+					   diff_rorp.get_sha1()), 2)
+
 	def patch_to_temp(self, basis_rp, diff_rorp, new):
 		"""Patch basis_rp, writing output in new, which doesn't exist yet"""
 		if diff_rorp.isflaglinked():
 			Hardlink.link_rp(diff_rorp, new, self.basis_root_rp)
-		elif diff_rorp.get_attached_filetype() == 'snapshot':
-			rpath.copy(diff_rorp, new)
+			return
+		if diff_rorp.get_attached_filetype() == 'snapshot':
+			copy_report = rpath.copy(diff_rorp, new)
 		else:
 			assert diff_rorp.get_attached_filetype() == 'diff'
-			Rdiff.patch_local(basis_rp, diff_rorp, new)
+			copy_report = Rdiff.patch_local(basis_rp, diff_rorp, new)
+		self.check_hash(copy_report, diff_rorp)
 		if new.lstat(): rpath.copy_attribs(diff_rorp, new)
 
 	def start_process(self, index, diff_rorp):
