@@ -82,7 +82,12 @@ def move(rpin, rpout):
 		rpin.delete()
 
 def copy(rpin, rpout, compress = 0):
-	"""Copy RPath rpin to rpout.  Works for symlinks, dirs, etc."""
+	"""Copy RPath rpin to rpout.  Works for symlinks, dirs, etc.
+
+	Returns close value of input for regular file, which can be used
+	to pass hashes on.
+
+	"""
 	log.Log("Regular copying %s to %s" % (rpin.index, rpout.path), 6)
 	if not rpin.lstat():
 		if rpout.lstat(): rpout.delete()
@@ -93,7 +98,7 @@ def copy(rpin, rpout, compress = 0):
 			rpout.delete()   # easier to write than compare
 		else: return
 
-	if rpin.isreg(): copy_reg_file(rpin, rpout, compress)
+	if rpin.isreg(): return copy_reg_file(rpin, rpout, compress)
 	elif rpin.isdir(): rpout.mkdir()
 	elif rpin.issym(): rpout.symlink(rpin.readlink())
 	elif rpin.ischardev():
@@ -111,11 +116,11 @@ def copy_reg_file(rpin, rpout, compress = 0):
 	try:
 		if (rpout.conn is rpin.conn and
 			rpout.conn is not Globals.local_connection):
-			rpout.conn.rpath.copy_reg_file(rpin.path, rpout.path, compress)
+			v = rpout.conn.rpath.copy_reg_file(rpin.path, rpout.path, compress)
 			rpout.setdata()
-			return
+			return v
 	except AttributeError: pass
-	rpout.write_from_fileobj(rpin.open("rb"), compress = compress)
+	return rpout.write_from_fileobj(rpin.open("rb"), compress = compress)
 
 def cmp(rpin, rpout):
 	"""True if rpin has the same data as rpout
@@ -158,7 +163,8 @@ def copy_attribs(rpin, rpout):
 	if (Globals.resource_forks_write and rpin.isreg() and
 		rpin.has_resource_fork()):
 		rpout.write_resource_fork(rpin.get_resource_fork())
-	if Globals.carbonfile_write and rpin.isreg():
+	if (Globals.carbonfile_write and rpin.isreg() and
+		rpin.has_carbonfile()):
 		rpout.write_carbonfile(rpin.get_carbonfile())
 	if Globals.eas_write: rpout.write_ea(rpin.get_ea())
 	rpout.chmod(rpin.getperms())
@@ -180,7 +186,8 @@ def copy_attribs_inc(rpin, rpout):
 	if (Globals.resource_forks_write and rpin.isreg() and
 		rpin.has_resource_fork() and rpout.isreg()):
 		rpout.write_resource_fork(rpin.get_resource_fork())
-	if Globals.carbonfile_write and rpin.isreg() and rpout.isreg():
+	if (Globals.carbonfile_write and rpin.isreg() and
+		rpin.has_carbonfile() and rpout.isreg()):
 		rpout.write_carbonfile(rpin.get_carbonfile())
 	if Globals.eas_write: rpout.write_ea(rpin.get_ea())
 	if rpin.isdir() and not rpout.isdir():
@@ -349,6 +356,7 @@ class RORPath:
 			elif key == 'carbonfile' and not Globals.carbonfile_write: pass
 			elif key == 'resourcefork' and not Globals.resource_forks_write:
 				pass
+			elif key == 'sha1': pass # one or other may not have set
 			elif (not other.data.has_key(key) or
 				  self.data[key] != other.data[key]): return 0
 
@@ -646,6 +654,18 @@ class RORPath:
 		"""Record resource fork in dictionary.  Does not write"""
 		self.data['resourcefork'] = rfork
 
+	def has_sha1(self):
+		"""True iff self has its sha1 digest set"""
+		return self.data.has_key('sha1')
+
+	def get_sha1(self):
+		"""Return sha1 digest.  Causes exception unless set_sha1 first"""
+		return self.data['sha1']
+
+	def set_sha1(self, digest):
+		"""Set sha1 hash (should be in hexdecimal)"""
+		self.data['sha1'] = digest
+
 
 class RPath(RORPath):
 	"""Remote Path class - wrapper around a possibly non-local pathname
@@ -978,16 +998,16 @@ class RPath(RORPath):
 		"""Reads fp and writes to self.path.  Closes both when done
 
 		If compress is true, fp will be gzip compressed before being
-		written to self.
+		written to self.  Returns closing value of fp.
 
 		"""
 		log.Log("Writing file object to " + self.path, 7)
 		assert not self.lstat(), "File %s already exists" % self.path
 		outfp = self.open("wb", compress = compress)
 		copyfileobj(fp, outfp)
-		if fp.close() or outfp.close():
-			raise RPathException("Error closing file")
+		if outfp.close(): raise RPathException("Error closing file")
 		self.setdata()
+		return fp.close()
 
 	def write_string(self, s, compress = None):
 		"""Write string s into rpath"""
@@ -1110,16 +1130,6 @@ class RPath(RORPath):
 		self.fsync(fp)
 		if Globals.fsync_directories: self.get_parent_rp().fsync()
 
-	def sync_delete(self):
-		"""Delete self with sync to guarantee completion
-
-		On some filesystems (like linux's ext2), we must sync both the
-		file and the directory to make sure.
-
-		"""
-		if self.lstat() and not self.issym(): self.fsync_local(self.delete)
-		if Globals.fsync_directories: self.get_parent_rp().fsync()
-
 	def get_data(self):
 		"""Open file as a regular file, read data, close, return data"""
 		fp = self.open("rb")
@@ -1153,36 +1163,9 @@ class RPath(RORPath):
 		ea.write_to_rp(self)
 		self.data['ea'] = ea
 
-	def get_carbonfile(self):
-		"""Return resource fork data, loading from filesystem if
-		necessary."""
-		from Carbon.File import FSSpec
-		import MacOS
-		try: return self.data['cfile']
-		except KeyError: pass
-
-		try:
-			fsobj = FSSpec(self.path)
-			finderinfo = fsobj.FSpGetFInfo()
-			cfile = {'creator': finderinfo.Creator,
-					 'type': finderinfo.Type,
-					 'location': finderinfo.Location,
-					 'flags': finderinfo.Flags}
-			self.data['carbonfile'] = cfile
-			return cfile
-		except MacOS.Error:
-			log.Log("Cannot read carbonfile information from %s" %
-					(self.path,), 2)
-			self.data['carbonfile'] = None 
-			return self.data['carbonfile']
-
 	def write_carbonfile(self, cfile):
 		"""Write new carbon data to self."""
-		if not cfile:
-			# This should be made cleaner---if you know Mac OS X tell
-			# me what could cause an error in get_carbonfile above
-			return
-			
+		if not cfile: return
 		log.Log("Writing carbon data to %s" % (self.index,), 7)
 		from Carbon.File import FSSpec
 		import MacOS
@@ -1248,7 +1231,25 @@ def setdata_local(rpath):
 	if Globals.acls_conn: rpath.data['acl'] = acl_get(rpath)
 	if Globals.resource_forks_conn and rpath.isreg():
 		rpath.get_resource_fork()
-	if Globals.carbonfile_conn and rpath.isreg(): rpath.get_carbonfile()
+	if Globals.carbonfile_conn and rpath.isreg():
+		rpath.data['carbonfile'] = carbonfile_get(rpath)
+
+def carbonfile_get(rpath):
+	"""Return carbonfile value for local rpath"""
+	from Carbon.File import FSSpec
+	import MacOS
+	try:
+		fsobj = FSSpec(self.path)
+		finderinfo = fsobj.FSpGetFInfo()
+		cfile = {'creator': finderinfo.Creator,
+				 'type': finderinfo.Type,
+				 'location': finderinfo.Location,
+				 'flags': finderinfo.Flags}
+		return cfile
+	except MacOS.Error:
+		log.Log("Cannot read carbonfile information from %s" %
+				(self.path,), 2)
+		return None
 
 
 # These functions are overwritten by the eas_acls.py module.  We can't
